#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Clients.Common;
using HeuristicLab.Common;
using HeuristicLab.Services.Hive.Common;
using HeuristicLab.Services.Hive.Common.DataTransfer;
using HeuristicLab.Services.Hive.Common.ServiceContracts;

namespace HeuristicLab.Clients.Hive.Slave {

  /// <summary>
  /// WcfService class is implemented as a Singleton and works as a communication Layer with the Server
  /// </summary>
  public class WcfService : MarshalByRefObject {
    private static WcfService instance;
    /// <summary>
    /// Getter for the Instance of the WcfService
    /// </summary>
    /// <returns>the Instance of the WcfService class</returns>
    public static WcfService Instance {
      get {
        if (instance == null) {
          instance = new WcfService();
        }
        return instance;
      }
    }

    public DateTime ConnectedSince { get; private set; }
    public NetworkEnum.WcfConnState ConnState { get; private set; }

    public event EventHandler Connected;
    private void OnConnected() {
      var handler = Connected;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler<EventArgs<Exception>> ExceptionOccured;
    private void OnExceptionOccured(Exception e) {
      var handler = ExceptionOccured;
      if (handler != null) handler(this, new EventArgs<Exception>(e));
    }

    /// <summary>
    /// Constructor
    /// </summary>
    private WcfService() {
      ConnState = NetworkEnum.WcfConnState.Disconnected;
    }

    /// <summary>
    /// Connects with the Server, registers the events and fires the Connected (and quiet possibly the ConnectionRestored) Event.
    /// </summary>
    public void Connect(HeuristicLab.Services.Hive.Common.DataTransfer.Slave slaveInfo) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          ConnState = NetworkEnum.WcfConnState.Connected;
          ConnectedSince = DateTime.Now;
          service.Obj.Hello(slaveInfo);
          OnConnected();
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
        }
      }
    }

    /// <summary>
    /// Disconnects the Slave from the Server
    /// </summary>
    public void Disconnect() {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          service.Obj.GoodBye(ConfigManager.Instance.GetClientInfo().Id);
          ConnState = NetworkEnum.WcfConnState.Disconnected;
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
        }
      }
    }

    /// <summary>
    /// Network communication Error Handler - Every network error gets logged and the connection switches to faulted state
    /// </summary>
    /// <param name="e">The Exception</param>
    private void HandleNetworkError(Exception e) {
      ConnState = NetworkEnum.WcfConnState.Failed;
      OnExceptionOccured(e);
    }

    /// <summary>
    /// Get a Job from the Server
    /// </summary>
    public Job GetJob(Guid jobId) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          Job job = service.Obj.GetJob(jobId);
          return job;
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
          return null;
        }
      }
    }

    public JobData GetJobData(Guid jobId) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          JobData jobData = service.Obj.GetJobData(jobId);
          return jobData;
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
          return null;
        }
      }
    }

    public void UpdateJob(Job job) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          service.Obj.UpdateJob(job);
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
        }
      }
    }

    /// <summary>
    /// used on pause or if job finished to upload the job results
    /// </summary>
    /// <param name="job"></param>
    /// <param name="jobData"></param>    
    public void UpdateJobData(Job job, JobData jobData, Guid slaveId) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          JobState before = job.State;
          job.SetState(JobState.Transferring, slaveId, "");
          service.Obj.UpdateJob(job);

          service.Obj.UpdateJobData(job, jobData);

          job.SetState(before, slaveId, "");
          service.Obj.UpdateJob(job);
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
        }
      }
    }
    
    public List<MessageContainer> SendHeartbeat(Heartbeat heartbeat) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          List<MessageContainer> msg = service.Obj.Heartbeat(heartbeat);
          return msg;
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
          return null;
        }
      }
    }

    public IEnumerable<PluginData> GetPluginDatas(List<Guid> pluginIds) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          IEnumerable<PluginData> msg = service.Obj.GetPluginDatas(pluginIds);
          return msg;
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
          return null;
        }
      }
    }

    public IEnumerable<Plugin> GetPlugins() {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          IEnumerable<Plugin> msg = service.Obj.GetPlugins();
          return msg;
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
          return null;
        }
      }
    }

    public Guid AddChildJob(Guid parentJobId, Job job, JobData jobData) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          Guid msg = service.Obj.AddChildJob(parentJobId, job, jobData);
          return msg;
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
          return new Guid();
        }
      }
    }

    public IEnumerable<JobData> GetChildJobs(Guid? parentJobId) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          IEnumerable<LightweightJob> msg = service.Obj.GetLightweightChildJobs(parentJobId, false, false);

          List<JobData> jobs = new List<JobData>();
          foreach (LightweightJob ljob in msg)
            jobs.Add(service.Obj.GetJobData(ljob.Id));

          return jobs;
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
          return null;
        }
      }
    }

    public void DeleteChildJobs(Guid jobId) {
      using (Disposable<IHiveService> service = GetSlaveService()) {
        try {
          service.Obj.DeleteChildJobs(jobId);
        }
        catch (Exception ex) {
          HandleNetworkError(ex);
        }
      }
    }

    private static Disposable<IHiveService> GetSlaveService() {
      return ServiceLocator.Instance.GetService("hiveslave", "hiveslave");
    }
  }
}